# -*- coding: utf-8 -*-
import os
import re
import builtins
import shutil
import zipfile
from os.path import isfile, join, isdir
import squish

from helpers.SetupClientHelper import get_resource_path, get_temp_resource_path
from helpers.SyncHelper import wait_for_client_to_be_ready
from helpers.ConfigHelper import get_config, is_owncloud_client
from helpers.FilesHelper import (
    build_conflicted_regex,
    sanitize_path,
    can_read,
    can_write,
    read_file_content,
    get_size_in_bytes,
    prefix_path_namespace,
    remember_path,
    write_file,
)
from helpers.UserHelper import get_username_for_user


def folder_exists(folder_path, timeout=1000):
    return squish.waitFor(
        lambda: isdir(sanitize_path(folder_path)),
        timeout,
    )


def file_exists(file_path, timeout=1000):
    return squish.waitFor(
        lambda: isfile(sanitize_path(file_path)),
        timeout,
    )


# To create folders in a temporary directory, we set is_temp_folder True
# And if is_temp_folder is True, the create_folder function create folders in tempFolderPath
def create_folder(foldername, username=None, is_temp_folder=False):
    if is_temp_folder:
        folder_path = join(get_config('tempFolderPath'), foldername)
    else:
        folder_path = get_resource_path(foldername, username)
    os.makedirs(prefix_path_namespace(folder_path))


def rename_file_folder(source, destination):
    source = get_resource_path(source)
    destination = get_resource_path(destination)
    os.rename(source, destination)


def create_file_with_size(filename, filesize, is_temp_folder=False):
    if is_temp_folder:
        file = join(get_config('tempFolderPath'), filename)
    else:
        file = get_resource_path(filename)
    with open(prefix_path_namespace(file), 'wb') as f:
        f.seek(get_size_in_bytes(filesize) - 1)
        f.write(b'\0')


def wait_and_write_file(path, content):
    wait_for_client_to_be_ready()
    write_file(path, content)


def wait_and_try_to_write_file(resource, content):
    wait_for_client_to_be_ready()
    try:
        write_file(resource, content)
    except:
        pass


def create_zip(resources, zip_file_name, cwd=''):
    os.chdir(cwd)
    with zipfile.ZipFile(zip_file_name, 'w') as zipped_file:
        for resource in resources:
            zipped_file.write(resource)


def extract_zip(zip_file_path, destination_dir):
    with zipfile.ZipFile(zip_file_path, 'r') as zip_file:
        zip_file.extractall(destination_dir)


def add_copy_suffix(resource_path, resource_type):
    if resource_type == 'file':
        source_dir = resource_path.rsplit('.', 1)
        return source_dir[0] + ' - Copy.' + source_dir[-1]
    return resource_path + ' - Copy'


@When(
    'user "|any|" creates a file "|any|" with the following content inside the sync folder'
)
def step(context, username, filename):
    if not is_owncloud_client():
        username = get_username_for_user(username)
    file_content = '\n'.join(context.multiLineText)
    file = get_resource_path(filename, username)
    wait_and_write_file(file, file_content)


@When('user "|any|" creates a folder "|any|" inside the sync folder')
def step(context, username, foldername):
    if not is_owncloud_client():
        username = get_username_for_user(username)
    wait_for_client_to_be_ready()
    create_folder(foldername, username)


@Given('user "|any|" has created a folder "|any|" inside the sync folder')
def step(context, username, foldername):
    if not is_owncloud_client():
        username = get_username_for_user(username)
    create_folder(foldername, username)


@When('user "|any|" creates a file "|any|" with size "|any|" inside the sync folder')
def step(context, _, filename, filesize):
    create_file_with_size(filename, filesize)


@When(r'the user copies the (file|folder) "([^"]*)" to "([^"]*)"', regexp=True)
def step(context, resource_type, source_dir, destination_dir):
    source = get_resource_path(source_dir)
    destination = get_resource_path(destination_dir)
    if source == destination and destination_dir != '/':
        destination = add_copy_suffix(source, resource_type)
    if resource_type == 'folder':
        return shutil.copytree(source, destination)
    return shutil.copy2(source, destination)


@When(r'the user renames a (?:file|folder) "([^"]*)" to "([^"]*)"', regexp=True)
def step(context, source, destination):
    wait_for_client_to_be_ready()
    rename_file_folder(source, destination)


@Then('the file "|any|" should exist on the file system with the following content')
def step(context, file_path):
    expected = '\n'.join(context.multiLineText)
    file_path = get_resource_path(file_path)
    with open(file_path, 'r', encoding='utf-8') as f:
        contents = f.read()
    test.compare(
        expected,
        contents.rstrip(),
        'file expected to exist with content '
        + expected
        + ' but does not have the expected content',
    )


@Then(r'^the (file|folder) "([^"]*)" should exist on the file system$', regexp=True)
def step(context, resource_type, resource):
    resource_path = get_resource_path(resource)
    resource_exists = False
    if resource_type == 'file':
        resource_exists = file_exists(
            resource_path, get_config('maxSyncTimeout') * 1000
        )
    else:
        resource_exists = folder_exists(
            resource_path, get_config('maxSyncTimeout') * 1000
        )

    test.compare(
        True,
        resource_exists,
        f'Assert {resource_type} "{resource}" exists on the system',
    )


@Then(r'^the (file|folder) "([^"]*)" should not exist on the file system$', regexp=True)
def step(context, resource_type, resource):
    resource_path = get_resource_path(resource)
    resource_exists = False
    if resource_type == 'file':
        resource_exists = file_exists(resource_path, 1000)
    else:
        resource_exists = folder_exists(resource_path, 1000)

    test.compare(
        False,
        resource_exists,
        f'Assert {resource_type} "{resource}" doesn\'t exist on the system',
    )


@Given('the user has changed the content of local file "|any|" to:')
def step(context, filename):
    file_content = '\n'.join(context.multiLineText)
    wait_and_write_file(get_resource_path(filename), file_content)


@Then(
    'a conflict file for "|any|" should exist on the file system with the following content'
)
def step(context, filename):
    expected = '\n'.join(context.multiLineText)

    onlyfiles = [
        f for f in os.listdir(get_resource_path()) if isfile(get_resource_path(f))
    ]
    found = False
    pattern = re.compile(build_conflicted_regex(filename))
    for file in onlyfiles:
        if pattern.match(file):
            with open(get_resource_path(file), 'r', encoding='utf-8') as f:
                if f.read() == expected:
                    found = True
                    break

    if not found:
        raise AssertionError('Conflict file not found with given name')


@When('the user overwrites the file "|any|" with content "|any|"')
def step(context, resource, content):
    resource = get_resource_path(resource)
    wait_and_write_file(resource, content)


@When('the user tries to overwrite the file "|any|" with content "|any|"')
def step(context, resource, content):
    resource = get_resource_path(resource)
    wait_and_try_to_write_file(resource, content)


@When('user "|any|" tries to overwrite the file "|any|" with content "|any|"')
def step(context, user, resource, content):
    resource = get_resource_path(resource, user)
    wait_and_try_to_write_file(resource, content)


@When(r'the user deletes the (file|folder) "([^"]*)"', regexp=True)
def step(context, item_type, resource):
    wait_for_client_to_be_ready()

    resource_path = sanitize_path(get_resource_path(resource))
    if item_type == 'file':
        os.remove(resource_path)
    else:
        shutil.rmtree(resource_path)


@When('user "|any|" creates the following files inside the sync folder:')
def step(context, username):
    if not is_owncloud_client():
        username = get_username_for_user(username)
    wait_for_client_to_be_ready()

    for row in context.table[1:]:
        file = get_resource_path(row[0], username)
        write_file(file, '')


@Given('the user has created a folder "|any|" in temp folder')
def step(context, folder_name):
    create_folder(folder_name, is_temp_folder=True)


@Given(
    'the user has created "|any|" files each of size "|any|" bytes inside folder "|any|" in temp folder'
)
def step(context, file_number, file_size, folder_name):
    current_sync_path = get_temp_resource_path(folder_name)
    if folder_exists(current_sync_path):
        file_size = builtins.int(file_size)
        for i in range(0, builtins.int(file_number)):
            file_name = f'file{i}.txt'
            create_file_with_size(join(current_sync_path, file_name), file_size, True)
    else:
        raise FileNotFoundError(
            f"Folder '{folder_name}' does not exist in the temp folder"
        )


@When(
    r'user "([^"]*)" moves (folder|file) "([^"]*)" from the temp folder into the sync folder',
    regexp=True,
)
def step(context, username, _, resource_name):
    if get_config('predefined_users'):
        username = get_username_for_user(username)
    source_dir = join(get_config('tempFolderPath'), resource_name)
    destination_dir = get_resource_path('/', username)
    shutil.move(source_dir, destination_dir)


@When(
    r'user "([^"]*)" moves (?:folder|file) "([^"]*)" to the temp folder',
    regexp=True,
)
def step(context, _, resource_name):
    source_dir = get_resource_path(resource_name)
    destination_dir = get_temp_resource_path(resource_name)
    shutil.move(source_dir, destination_dir)


@When(
    r'user "([^"]*)" moves (?:file|folder) "([^"]*)" to "([^"]*)" in the sync folder',
    regexp=True,
)
def step(context, username, source, destination):
    wait_for_client_to_be_ready()
    source_dir = get_resource_path(source, username)
    if destination in (None, '/'):
        destination = ''
    destination_dir = get_resource_path(destination, username)
    shutil.move(source_dir, destination_dir)


@Then('user "|any|" should be able to open the file "|any|" on the file system')
def step(context, user, file_name):
    if not is_owncloud_client():
        user = get_username_for_user(user)
    file_path = get_resource_path(file_name, user)
    if not os.path.exists(file_path):
        raise ValueError(f'Could not find resource {file_path}')
    test.compare(can_read(file_path), True, 'File should be readable')


@Then('as "|any|" the file "|any|" should have content "|any|" on the file system')
def step(context, user, file_name, content):
    if not is_owncloud_client():
        user = get_username_for_user(user)
    file_path = get_resource_path(file_name, user)
    file_content = read_file_content(file_path)
    test.compare(file_content, content, 'Comparing file content')


@Then('user "|any|" should not be able to edit the file "|any|" on the file system')
def step(context, user, file_name):
    if not is_owncloud_client():
        user = get_username_for_user(user)
    file_path = get_resource_path(file_name, user)
    if not os.path.exists(file_path):
        raise ValueError(f'Could not find resource {file_path}')
    test.compare(not can_write(file_path), True, 'File should not be writable')


@Given(
    'the user has created a zip file "|any|" with the following resources in the temp folder'
)
def step(context, zip_file_name):
    resource_list = []

    for row in context.table[1:]:
        resource_list.append(row[0])
        resource = join(get_config('tempFolderPath'), row[0])
        if row[1] == 'folder':
            os.makedirs(resource)
        elif row[1] == 'file':
            content = ''
            if len(row) > 2 and row[2]:
                content = row[2]
            write_file(resource, content)
    create_zip(resource_list, zip_file_name, get_config('tempFolderPath'))


@When('user "|any|" unzips the zip file "|any|" inside the sync root')
def step(context, username, zip_file_name):
    if get_config('predefined_users'):
        username = get_username_for_user(username)
    destination_dir = get_resource_path('/', username)
    zip_file_path = join(destination_dir, zip_file_name)
    extract_zip(zip_file_path, destination_dir)


@When('user "|any|" copies file "|any|" to temp folder')
def step(context, username, source):
    wait_for_client_to_be_ready()
    source_dir = get_resource_path(source, username)
    destination_dir = get_temp_resource_path(source)
    shutil.copy2(source_dir, destination_dir)


@Given('the user has created folder "|any|" in the default home path')
def step(context, folder_name):
    folder_path = join(get_config('home_dir'), folder_name)
    os.makedirs(prefix_path_namespace(folder_path))
    remember_path(folder_path)
    # when account is added, folder with suffix will be created
    remember_path(f'{folder_path} (2)')



import os
import glob
import shutil
import test
import squish
import squishinfo

from helpers.ConfigHelper import get_config
from helpers.FilesHelper import prefix_path_namespace


def get_screenrecords_path():
    return os.path.join(get_config("guiTestReportDir"), "screenrecords")


def get_screenshots_path():
    return os.path.join(get_config("guiTestReportDir"), "screenshots")


def is_video_enabled():
    return get_config("record_video_on_failure") and not reached_video_limit()


def reached_video_limit():
    video_report_dir = get_screenrecords_path()
    if not os.path.exists(video_report_dir):
        return False
    entries = [f for f in os.scandir(video_report_dir) if f.is_file()]
    return len(entries) >= get_config("video_record_limit")


def save_video_recording(filename, test_failed):
    try:
        # do not throw if stopVideoCapture() fails
        test.stopVideoCapture()
    except:
        test.log("Failed to stop screen recording")

    if not (video_dir := squishinfo.resultDir):
        video_dir = squishinfo.testCase
    else:
        test_case = "/".join(squishinfo.testCase.split("/")[-2:])
        video_dir = os.path.join(video_dir, test_case)
    video_dir = os.path.join(video_dir, "attachments")

    # if the test failed
    # move videos to the screenrecords directory
    if test_failed:
        video_files = glob.glob(f"{video_dir}/**/*.mp4", recursive=True)
        screenrecords_dir = get_screenrecords_path()
        if not os.path.exists(screenrecords_dir):
            os.makedirs(screenrecords_dir)
        # reverse the list to get the latest video first
        video_files.reverse()
        for idx, video in enumerate(video_files):
            if idx:
                file_parts = filename.rsplit(".", 1)
                filename = f"{file_parts[0]}_{idx+1}.{file_parts[1]}"
            shutil.move(video, os.path.join(screenrecords_dir, filename))
    # remove the video directory
    shutil.rmtree(prefix_path_namespace(video_dir))


def take_screenshot(filename):
    directory = get_screenshots_path()
    if not os.path.exists(directory):
        os.makedirs(directory)
    try:
        squish.saveDesktopScreenshot(os.path.join(directory, filename))
    except:
        test.log("Failed to save screenshot")



import os
import re
import sys
import test
import squish
import helpers.api.http_helper as request

from helpers.ConfigHelper import get_config, is_linux, is_windows
from helpers.FilesHelper import sanitize_path

if is_windows():
    from helpers.WinPipeHelper import WinPipeConnect as SocketConnect
else:
    # NOTE: 'syncstate.py' was removed from client
    # and is now available at https://github.com/owncloud/client-desktop-shell-integration-nautilus
    # check if 'syncstate.py' is available, if not, download it
    custom_lib = get_config('custom_lib')
    syncstate_lib_file = os.path.join(custom_lib, 'syncstate.py')
    os.makedirs(custom_lib, exist_ok=True)

    if not os.path.exists(syncstate_lib_file):
        response = request.get(
            (
                'https://raw.githubusercontent.com/'
                'owncloud/client-desktop-shell-integration-nautilus/master/src/syncstate.py'
            )
        )
        assert response.status_code == 200, 'Failed to get content of syncstate file'
        with open(syncstate_lib_file, 'w', encoding='utf-8') as f:
            f.write(response.text)

    # the script needs to use the system-wide python
    # to switch from the built-in interpreter
    #   see https://kb.froglogic.com/squish/howto/using-external-python-interpreter-squish-6-6/
    # if the IDE fails to reference the script,
    # add the folder in Edit->Preferences->PyDev->Interpreters->Libraries
    sys.path.append(custom_lib)
    from custom_lib.syncstate import SocketConnect

# socket messages
socket_messages = []
SOCKET_CONNECT = None
# Whether wait has been made or not after account is set up
# This is useful for waiting only for the first time
WAITED_AFTER_SYNC = False

# File syncing in client has the following status
SYNC_STATUS = {
    'SYNC': 'STATUS:SYNC',  # sync in progress
    'OK': 'STATUS:OK',  # sync completed
    'ERROR': 'STATUS:ERROR',  # sync error
    'IGNORE': 'STATUS:IGNORE',  # sync ignored
    'NOP': 'STATUS:NOP',  # not in sync yet
    'REGISTER': 'REGISTER_PATH',
    'UNREGISTER': 'UNREGISTER_PATH',
    'UPDATE': 'UPDATE_VIEW',
}

SYNC_PATTERNS = {
    # default sync patterns for the initial sync (after adding account)
    # the pattern can be of TWO types depending on the available resources (files/folders)
    'initial': [
        # when adding account via New Account wizard
        [
            SYNC_STATUS['REGISTER'],
            SYNC_STATUS['UPDATE'],
            SYNC_STATUS['UPDATE'],
        ],
        # when syncing empty account (hidden files are ignored)
        [SYNC_STATUS['UPDATE'], SYNC_STATUS['OK']],
        # when syncing an account that has some files/folders
        [SYNC_STATUS['SYNC'], SYNC_STATUS['OK']],
    ],
    'root_synced': [
        [
            SYNC_STATUS['SYNC'],
            SYNC_STATUS['OK'],
            SYNC_STATUS['OK'],
            SYNC_STATUS['OK'],
            SYNC_STATUS['UPDATE'],
        ],
        [
            SYNC_STATUS['SYNC'],
            SYNC_STATUS['UPDATE'],
            SYNC_STATUS['OK'],
            SYNC_STATUS['OK'],
            SYNC_STATUS['OK'],
            SYNC_STATUS['UPDATE'],
        ],
    ],
    'single_synced': [SYNC_STATUS['SYNC'], SYNC_STATUS['OK']],
    'error': [SYNC_STATUS['ERROR']],
}


def get_socket_connection():
    global SOCKET_CONNECT
    if not SOCKET_CONNECT or not SOCKET_CONNECT.connected:
        SOCKET_CONNECT = SocketConnect()
    return SOCKET_CONNECT


def read_socket_messages():
    messages = []
    socket_connect = get_socket_connection()
    socket_connect.read_socket_data_with_timeout(0.1)
    for line in socket_connect.get_available_responses():
        messages.append(line)
    return messages


def read_and_update_socket_messages():
    messages = read_socket_messages()
    return update_socket_messages(messages)


def update_socket_messages(messages):
    socket_messages.extend(filter_sync_messages(messages))
    return socket_messages


def clear_socket_messages(resource=''):
    global socket_messages
    if resource:
        resource_messages = set(filter_messages_for_item(socket_messages, resource))
        socket_messages = [
            msg for msg in socket_messages if msg not in resource_messages
        ]
    else:
        socket_messages.clear()


def close_socket_connection():
    socket_messages.clear()
    if SOCKET_CONNECT:
        SOCKET_CONNECT.connected = False
        if is_windows():
            SOCKET_CONNECT.close_conn()
        elif is_linux():
            SOCKET_CONNECT._sock.close()  # pylint: disable=protected-access


def get_initial_sync_patterns():
    return SYNC_PATTERNS['initial']


def get_synced_pattern(resource=''):
    # get only the resource path
    sync_path = get_config('currentUserSyncPath')
    if get_config('ocis'):
        sync_path = os.path.join(sync_path, get_config('syncConnectionName'))

    if resource := resource.replace(sync_path, '').strip('\\').strip('/'):
        return SYNC_PATTERNS['single_synced']
    return SYNC_PATTERNS['root_synced']


# generate sync pattern from the socket messages
#
# returns List
# e.g: ['UPDATE_VIEW', 'STATUS:OK']
def generate_sync_pattern_from_messages(messages):
    pattern = []
    if not messages:
        return pattern

    sync_messages = filter_sync_messages(messages)
    for message in sync_messages:
        # E.g; from "STATUS:OK:/tmp/client-bdd/Alice/"
        # excludes ":/tmp/client-bdd/Alice/"
        # adds only "STATUS:OK" to the pattern list
        if match := re.search(':(/|[A-Z]{1}:\\\\).*', message):
            (end, _) = match.span()
            # shared resources will have status like "STATUS:OK+SWM"
            status = message[:end].replace('+SWM', '')
            pattern.append(status)
    return pattern


# strip out the messages that are not related to sync
def filter_sync_messages(messages):
    start_idx = 0
    if 'GET_STRINGS:END' in messages:
        start_idx = messages.index('GET_STRINGS:END') + 1
    return messages[start_idx:]


def filter_messages_for_item(messages, item):
    filtered_messages = []
    for msg in messages:
        msg = msg.rstrip('/').rstrip('\\')
        item = item.rstrip('/').rstrip('\\')
        if msg.endswith(item):
            filtered_messages.append(msg)
    return filtered_messages


def listen_sync_status_for_item(item, resource_type='FOLDER'):
    if (resource_type := resource_type.upper()) not in ('FILE', 'FOLDER'):
        raise ValueError('resource_type must be "FILE" or "FOLDER"')
    socket_connect = get_socket_connection()
    item = item.rstrip('\\')
    socket_connect.sendCommand(f'RETRIEVE_{resource_type}_STATUS:{item}\n')


def get_current_sync_status(resource, resource_type):
    listen_sync_status_for_item(resource, resource_type)
    messages = filter_messages_for_item(read_socket_messages(), resource)
    # return the last message from the list
    return messages[-1]


def wait_for_resource_to_sync(resource, resource_type='FOLDER', patterns=None):
    listen_sync_status_for_item(resource, resource_type)

    timeout = get_config('maxSyncTimeout') * 1000

    if patterns is None:
        patterns = get_synced_pattern(resource)

    synced = squish.waitFor(
        lambda: has_sync_pattern(patterns, resource),
        timeout,
    )
    clear_socket_messages(resource)
    if not synced:
        # if the sync pattern doesn't match then check the last sync status
        # and pass the step if the last sync status is STATUS:OK
        status = get_current_sync_status(resource, resource_type)
        if status.startswith(SYNC_STATUS['OK']):
            test.log(
                '[WARN] Failed to match sync pattern for resource: '
                + resource
                + f'\nBut its last status is "{SYNC_STATUS["OK"]}"'
                + '. So passing the step.'
            )
            return
        raise TimeoutError(
            'Timeout while waiting for sync to complete for '
            + str(timeout)
            + ' milliseconds'
        )


def wait_for_initial_sync_to_complete(path):
    wait_for_resource_to_sync(
        path,
        'FOLDER',
        get_initial_sync_patterns(),
    )


def has_sync_pattern(patterns, resource=None):
    if isinstance(patterns[0], str):
        patterns = [patterns]
    messages = read_and_update_socket_messages()
    if resource:
        messages = filter_messages_for_item(messages, resource)
    for pattern in patterns:
        pattern_len = len(pattern)
        for idx, _ in enumerate(messages):
            actual_pattern = generate_sync_pattern_from_messages(
                messages[idx : idx + pattern_len]
            )
            if len(actual_pattern) < pattern_len:
                break
            if pattern_len == len(actual_pattern) and pattern == actual_pattern:
                return True
    # 100 milliseconds polling interval
    squish.snooze(0.1)
    return False


# Using socket API to check file sync status
def has_sync_status(item_name, status):
    sync_messages = read_and_update_socket_messages()
    sync_messages = filter_messages_for_item(sync_messages, item_name)
    for line in sync_messages:
        line = line.rstrip('/').rstrip('\\')
        item_name = item_name.rstrip('/').rstrip('\\')
        if line.startswith(status) and line.endswith(item_name):
            return True
    return False


# useful for checking sync status such as 'error', 'ignore'
# but not quite so reliable for checking 'ok' sync status
def wait_for_resource_to_have_sync_status(
    resource, resource_type, status=SYNC_STATUS['OK'], timeout=None
):
    resource = sanitize_path(resource)

    listen_sync_status_for_item(resource, resource_type)

    if not timeout:
        timeout = get_config('maxSyncTimeout') * 1000

    result = squish.waitFor(
        lambda: has_sync_status(resource, status),
        timeout,
    )

    if not result:
        if status == SYNC_STATUS['ERROR']:
            expected = 'have sync error'
        elif status == SYNC_STATUS['IGNORE']:
            expected = 'be sync ignored'
        else:
            expected = 'be synced'
        raise ValueError(
            f'Expected {resource_type} "{resource}" to {expected}, but not.'
        )


def wait_for_resource_to_have_sync_error(resource, resource_type):
    wait_for_resource_to_have_sync_status(resource, resource_type, SYNC_STATUS['ERROR'])


# performing actions immediately after completing the sync from the server does not work
# The test should wait for a while before performing the action
# issue: https://github.com/owncloud/client/issues/8832
def wait_for_client_to_be_ready():
    global WAITED_AFTER_SYNC
    if not WAITED_AFTER_SYNC:
        squish.snooze(get_config('minSyncTimeout'))
        WAITED_AFTER_SYNC = True


def clear_waited_after_sync():
    global WAITED_AFTER_SYNC
    WAITED_AFTER_SYNC = False



import os
import time

# pylint: disable=import-error
import win32pipe
import win32file
import winerror
import pywintypes
import win32event

TIMEOUT = 100
DEFAULT_BUFLEN = 4096
CLIENT_MESSAGES = [
    'STATUS',
    'REGISTER_PATH',
    'UNREGISTER_PATH',
    'UPDATE_VIEW',
    'GET_STRINGS',
    'STRING',
    'VERSION',
    'SHARE',
    'GET_MENU_ITEMS',
    'MENU_ITEM',
]


def get_pipe_path():
    pipename = r'\\.\\pipe\\'
    pipename = os.path.join(pipename, 'ownCloud-' + os.getenv('USERNAME'))
    return pipename


class WinPipeConnect:
    def __init__(self):
        self.connected = False
        self._pipe = None
        self._remainder = ''.encode('utf-8')
        self._overlapped = pywintypes.OVERLAPPED()
        self._overlapped.hEvent = win32event.CreateEvent(None, 1, 0, None)
        self.connect_to_pipe_server()

    def connect_to_pipe_server(self):
        try:
            pipename = get_pipe_path()

            self._pipe = win32file.CreateFile(
                pipename,
                win32file.GENERIC_READ | win32file.GENERIC_WRITE,
                0,
                None,
                win32file.OPEN_EXISTING,
                win32file.FILE_FLAG_OVERLAPPED,
                None,
            )

            if self._pipe == win32file.INVALID_HANDLE_VALUE:
                win32pipe.CloseHandle(self._pipe)
                raise OSError('Invalid _pipe value')

            self.connected = True

            self.sendCommand('VERSION:\n')
            self.sendCommand('GET_STRINGS:\n')
        except Exception as e:  # pylint: disable=broad-except
            print(f'Could not connect to named pipe {pipename}\n' + str(e))
            win32file.CloseHandle(self._pipe)

    def sendCommand(self, cmd):  # pylint: disable=invalid-name
        if self.connected:
            w_res, _ = win32file.WriteFile(
                self._pipe, cmd.encode('utf-8'), self._overlapped
            )
            if w_res == winerror.ERROR_IO_PENDING:
                res = win32event.WaitForSingleObject(self._overlapped.hEvent, TIMEOUT)
                if res != win32event.WAIT_OBJECT_0:
                    print('Sending timed out!')
                    return False
                if not win32file.GetOverlappedResult(
                    self._pipe, self._overlapped, False
                ):
                    print('GetOverlappedResult failed')
                    return False
        else:
            print('Cannot send, not connected!')
            return False
        return True

    # Reads data that becomes available.
    # New responses can be accessed with get_available_responses().
    def read_socket_data_with_timeout(self, timeout):
        messages = b''
        start_time = time.time()
        while True:  # pylint: disable=too-many-nested-blocks
            if (time.time() - start_time) >= timeout:
                self._remainder += messages
                break

            peek_bytes = win32pipe.PeekNamedPipe(self._pipe, DEFAULT_BUFLEN)[1]
            if isinstance(peek_bytes, int) and peek_bytes > 0:
                _, message_mem = win32file.ReadFile(
                    self._pipe, DEFAULT_BUFLEN, self._overlapped
                )
                if message_mem:
                    m_bytes = bytes(message_mem)
                    if b'\n' in m_bytes:
                        for m in m_bytes.split(b'\n'):
                            try:
                                # append decodable and client specific bytes
                                m.decode('utf-8')
                                start = m.split(b':', 1)[0]
                                if start.decode('utf-8') in CLIENT_MESSAGES:
                                    messages += m + b'\n'
                            except:
                                pass

            else:
                res = win32event.WaitForSingleObject(
                    self._overlapped.hEvent, int(timeout * 1000)
                )
                if res != win32event.WAIT_OBJECT_0:
                    print('Reading timed out!')
                    return False
                if not win32file.GetOverlappedResult(
                    self._pipe, self._overlapped, False
                ):
                    return False
        return True

    # Parses response lines out of collected data, returns list of strings
    def get_available_responses(self):
        if (end := self._remainder.rfind(b'\n')) == -1:
            return []
        data = self._remainder[:end]
        self._remainder = self._remainder[end + 1 :]
        return data.decode('utf-8').split('\n')

    def close_conn(self):
        win32file.CloseHandle(self._pipe)



from urllib.parse import quote
import xml.etree.ElementTree as ET

import helpers.api.http_helper as request
from helpers.api.utils import url_join
from helpers.ConfigHelper import get_config
from helpers.FilesHelper import get_file_for_upload


def get_webdav_url():
    return url_join(get_config('localBackendUrl'), 'remote.php/dav/files')


def get_resource_path(user, resource):
    resource = resource.strip('/').replace('\\', '/')
    encoded_resource_path = [quote(path, safe='') for path in resource.split('/')]
    encoded_resource_path = '/'.join(encoded_resource_path)
    url = url_join(get_webdav_url(), user, encoded_resource_path)
    return url


def resource_exists(user, resource):
    response = request.propfind(get_resource_path(user, resource), user=user)
    if response.status_code == 207:
        return True
    if response.status_code == 404:
        return False
    raise AssertionError(f'Server returned status code: {response.status_code}')


def get_file_content(user, resource):
    response = request.get(get_resource_path(user, resource), user=user)
    return response.text


def get_folder_items_count(user, folder_name):
    folder_name = folder_name.strip('/')
    path = get_resource_path(user, folder_name)
    xml_response = request.propfind(path, user=user)
    total_items = 0
    root_element = ET.fromstring(xml_response.content)
    for response_element in root_element:
        for href_element in response_element:
            # The first item is folder itself so excluding it
            if href_element.tag == '{DAV:}href' and not href_element.text.endswith(
                f'{user}/{folder_name}/'
            ):
                total_items += 1
    return str(total_items)


def create_folder(user, folder_name):
    url = get_resource_path(user, folder_name)
    response = request.mkcol(url, user=user)
    assert (
        response.status_code == 201
    ), f'Could not create the folder: {folder_name} for user {user}'


def create_file(user, file_name, contents):
    url = get_resource_path(user, file_name)
    response = request.put(url, body=contents, user=user)
    assert response.status_code in [
        201,
        204,
    ], f"Could not create file '{file_name}' for user {user}"


def upload_file(user, file_name, destination):
    file_path = get_file_for_upload(file_name)
    with open(file_path, 'rb') as file:
        contents = file.read()
    create_file(user, destination, contents)


def delete_resource(user, resource):
    url = get_resource_path(user, resource)
    response = request.delete(url, user=user)
    assert response.status_code == 204, f"Could not delete folder '{resource}'"



import names
import squish
import object  # pylint: disable=redefined-builtin

from helpers.ConfigHelper import get_config


class SyncConnection:
    FOLDER_SYNC_CONNECTION_LIST = {
        "container": names.quickWidget_scrollView_ScrollView,
        "type": "ListView",
        "visible": True,
    }
    FOLDER_SYNC_CONNECTION = {
        "container": names.settings_stack_QStackedWidget,
        "name": "_folderList",
        "type": "QListView",
        "visible": 1,
    }
    FOLDER_SYNC_CONNECTION_LABEL = {
        "container": names.quickWidget_scrollView_ScrollView,
        "type": "Label",
        "unnamed": 1,
        "visible": True,
    }
    FOLDER_SYNC_CONNECTION_MENU_BUTTON = {
        "checkable": False,
        "container": names.quickWidget_scrollView_ScrollView,
        "type": "Button",
        "unnamed": 1,
        "visible": True,
    }
    MENU = {
        "type": "QMenu",
        "window": names.settings_OCC_SettingsDialog,
        "visible": 1,
    }
    DISABLE_VFS_CONFIRMATION_BUTTON = {
        "text": "Disable support",
        "type": "QPushButton",
        "visible": 1,
        "window": names.disable_virtual_file_support_QMessageBox,
    }
    SELECTIVE_SYNC_APPLY_BUTTON = {
        "container": names.settings_stack_QStackedWidget,
        "name": "selectiveSyncApply",
        "type": "QPushButton",
        "visible": 1,
    }
    CANCEL_FOLDER_SYNC_CONNECTION_DIALOG = {
        "text": "Cancel",
        "type": "QPushButton",
        "unnamed": 1,
        "visible": 1,
        "window": names.confirm_Folder_Sync_Connection_Removal_QMessageBox,
    }
    REMOVE_FOLDER_SYNC_CONNECTION_BUTTON = {
        "text": "Remove Folder Sync Connection",
        "type": "QPushButton",
        "unnamed": 1,
        "visible": 1,
        "window": names.confirm_Folder_Sync_Connection_Removal_QMessageBox,
    }

    @staticmethod
    def open_menu(sync_folder=""):
        if get_config("ocis") and not sync_folder:
            sync_folder = get_config("personal_sync_folder")
        elif not sync_folder:
            sync_folder = "ownCloud"
        selector = SyncConnection.FOLDER_SYNC_CONNECTION_LABEL.copy()
        selector.update({"text": sync_folder})

        menu_button = None

        # get the parent of the sync folder label
        parent = object.parent(
            squish.waitForObject(selector)
        ).parent.parent.parent.parent.parent
        children = object.children(squish.waitForObject(parent))
        for obj in children:
            # get sync folder menu button
            if obj.id == "moreButton":
                menu_button = squish.waitForObject(obj)
        squish.mouseClick(menu_button)

    @staticmethod
    def perform_action(action, sync_folder=""):
        SyncConnection.open_menu(sync_folder)
        squish.activateItem(squish.waitForObjectItem(SyncConnection.MENU, action))

    @staticmethod
    def force_sync():
        SyncConnection.perform_action("Force sync now")

    @staticmethod
    def pause_sync():
        SyncConnection.perform_action("Pause sync")

    @staticmethod
    def resume_sync():
        SyncConnection.perform_action("Resume sync")

    @staticmethod
    def enable_vfs():
        SyncConnection.perform_action("Enable virtual file support")

    @staticmethod
    def disable_vfs():
        SyncConnection.perform_action("Disable virtual file support")
        squish.clickButton(
            squish.waitForObject(SyncConnection.DISABLE_VFS_CONFIRMATION_BUTTON)
        )

    @staticmethod
    def has_menu_item(item):
        return squish.waitForObjectItem(SyncConnection.MENU, item)

    @staticmethod
    def menu_item_exists(menu_item):
        obj = SyncConnection.MENU.copy()
        obj.update({"type": "QAction", "text": menu_item})
        return object.exists(obj)

    @staticmethod
    def choose_what_to_sync():
        SyncConnection.open_menu()
        SyncConnection.perform_action("Choose what to sync")

    @staticmethod
    def unselect_folder_in_selective_sync(folder_name):
        sync_folders = object.children(
            squish.waitForObject(SyncConnection.FOLDER_SYNC_CONNECTION)
        )
        for sync_folder in sync_folders:
            # TODO: allow selective sync in other sync folders as well
            if hasattr(sync_folder, "text") and sync_folder.text == "Personal":
                items = object.children(sync_folder)
                for item in items:
                    if hasattr(item, "text") and item.text:
                        # remove item size suffix
                        # example: folder1 (13 B) => folder1
                        item_name = item.text.rsplit(" ", 1)
                        if item_name[0] == folder_name:
                            # NOTE: checkbox does not have separate object
                            # click on (11,11) which is a checkbox to unselect the folder
                            squish.mouseClick(
                                item,
                                11,
                                11,
                                squish.Qt.NoModifier,
                                squish.Qt.LeftButton,
                            )
                            break
        squish.clickButton(
            squish.waitForObject(SyncConnection.SELECTIVE_SYNC_APPLY_BUTTON)
        )

    @staticmethod
    def get_folder_connection_count():
        return squish.waitForObject(SyncConnection.FOLDER_SYNC_CONNECTION_LIST).count

    @staticmethod
    def remove_folder_sync_connection(sync_folder=""):
        SyncConnection.perform_action("Remove folder sync connection", sync_folder)

    @staticmethod
    def cancel_folder_sync_connection_removal():
        squish.clickButton(
            squish.waitForObject(SyncConnection.CANCEL_FOLDER_SYNC_CONNECTION_DIALOG)
        )

    @staticmethod
    def confirm_folder_sync_connection_removal():
        squish.clickButton(
            squish.waitForObject(SyncConnection.REMOVE_FOLDER_SYNC_CONNECTION_BUTTON)
        )



from datetime import datetime
import test
import names
import squish
import object  # pylint: disable=redefined-builtin


class PublicLinkDialog:
    PUBLIC_LINKS_TAB = {
        "container": names.sharingDialog_qt_tabwidget_tabbar_QTabBar,
        "text": "Public Links",
        "type": "TabItem",
    }
    PASSWORD_CHECKBOX = {
        "container": names.qt_tabwidget_stackedwidget_OCC_ShareLinkWidget_OCC_ShareLinkWidget,
        "name": "checkBox_password",
        "type": "QCheckBox",
        "visible": 1,
    }
    PASSWORD_INPUT_FIELD = {
        "container": names.qt_tabwidget_stackedwidget_OCC_ShareLinkWidget_OCC_ShareLinkWidget,
        "name": "lineEdit_password",
        "type": "QLineEdit",
        "visible": 1,
    }
    EXPIRYDATE_CHECKBOX = {
        "container": names.qt_tabwidget_stackedwidget_OCC_ShareLinkWidget_OCC_ShareLinkWidget,
        "name": "checkBox_expire",
        "type": "QCheckBox",
        "visible": 1,
    }
    CREATE_SHARE_BUTTON = {
        "container": names.qt_tabwidget_stackedwidget_OCC_ShareLinkWidget_OCC_ShareLinkWidget,
        "name": "createShareButton",
        "type": "QPushButton",
        "visible": 1,
    }
    PUBLIC_LINK_NAME = {
        "column": 0,
        "container": names.oCC_ShareLinkWidget_linkShares_QTableWidget,
        "row": 0,
        "type": "QModelIndex",
    }
    EXPIRATION_DATE_FIELD = {
        "container": names.qt_tabwidget_stackedwidget_OCC_ShareLinkWidget_OCC_ShareLinkWidget,
        "name": "qt_spinbox_lineedit",
        "type": "QLineEdit",
        "visible": 1,
    }
    READ_ONLY_RADIO_BUTTON = {
        "container": names.qt_tabwidget_stackedwidget_OCC_ShareLinkWidget_OCC_ShareLinkWidget,
        "name": "radio_readOnly",
        "type": "QRadioButton",
        "visible": 1,
    }
    READ_WRITE_RADIO_BUTTON = {
        "container": names.qt_tabwidget_stackedwidget_OCC_ShareLinkWidget_OCC_ShareLinkWidget,
        "name": "radio_readWrite",
        "type": "QRadioButton",
        "visible": 1,
    }
    UPLOAD_ONLY_RADIO_BUTTON = {
        "container": names.qt_tabwidget_stackedwidget_OCC_ShareLinkWidget_OCC_ShareLinkWidget,
        "name": "radio_uploadOnly",
        "type": "QRadioButton",
        "visible": 1,
    }
    DELETE_LINK_BUTTON = {
        "container": names.oCC_ShareLinkWidget_linkShares_QTableWidget,
        "type": "QToolButton",
        "occurrence": 2,
    }
    CONFIRM_LINK_DELETE_BUTTON = {
        "container": names.qt_tabwidget_stackedwidget_OCC_ShareLinkWidget_OCC_ShareLinkWidget,
        "text": "Delete",
        "type": "QPushButton",
    }
    UPDATE_PASSWORD_BUTTON = {
        "container": names.qt_tabwidget_stackedwidget_OCC_ShareLinkWidget_OCC_ShareLinkWidget,
        "name": "pushButton_setPassword",
        "type": "QPushButton",
        "visible": 1,
    }

    DATE_FORMATS = [
        "%m/%d/%Y",
        "%d/%m/%Y",
        "%Y-%m-%d",
        "%d-%m-%Y",
        "%Y/%m/%d",
        "%m/%d/%y",
        "%d/%m/%y",
    ]
    # to store current default public link expiry date
    default_expiry_date = ""

    @staticmethod
    def parse_date(date):
        for date_format in PublicLinkDialog.DATE_FORMATS:
            try:
                date = str(datetime.strptime(date, date_format))
                date = date.split(" ", maxsplit=1)[0]
                break
            except:
                pass
        return date

    @staticmethod
    def set_default_expiry_date(default_date):
        PublicLinkDialog.default_expiry_date = PublicLinkDialog.parse_date(default_date)

    @staticmethod
    def get_default_expiry_date():
        return PublicLinkDialog.default_expiry_date

    @staticmethod
    def open_public_link_tab():
        squish.mouseClick(squish.waitForObject(PublicLinkDialog.PUBLIC_LINKS_TAB))

    @staticmethod
    def create_public_link(password="", permissions="", expire_date=""):
        permission_locator = ""
        if permissions:
            permission_locator = PublicLinkDialog.get_radio_object_for_permssion(
                permissions
            )

        if permission_locator:
            squish.clickButton(squish.waitForObject(permission_locator))

        if password:
            PublicLinkDialog.set_password(password)

        if expire_date:
            PublicLinkDialog.set_expiration_date(expire_date)

        squish.clickButton(squish.waitForObject(PublicLinkDialog.CREATE_SHARE_BUTTON))
        squish.waitFor(
            lambda: (
                squish.waitForObject(PublicLinkDialog.PUBLIC_LINK_NAME).displayText
                == "Public link"
            )
        )

    @staticmethod
    def toggle_password():
        squish.clickButton(squish.waitForObject(PublicLinkDialog.PASSWORD_CHECKBOX))

    @staticmethod
    def set_password(password):
        if not squish.waitForObjectExists(PublicLinkDialog.PASSWORD_CHECKBOX).checked:
            PublicLinkDialog.toggle_password()

        squish.mouseClick(squish.waitForObject(PublicLinkDialog.PASSWORD_INPUT_FIELD))
        squish.type(
            squish.waitForObject(PublicLinkDialog.PASSWORD_INPUT_FIELD),
            password,
        )

    @staticmethod
    def toggle_expiration_date():
        squish.clickButton(squish.waitForObject(PublicLinkDialog.EXPIRYDATE_CHECKBOX))

    @staticmethod
    def set_expiration_date(expire_date):
        enabled = squish.waitForObjectExists(
            PublicLinkDialog.EXPIRYDATE_CHECKBOX
        ).checked
        if not enabled:
            PublicLinkDialog.toggle_expiration_date()

        if not expire_date == "default":
            exp_date = datetime.strptime(expire_date, "%Y-%m-%d")
            exp_year = exp_date.year - 2000
            squish.mouseClick(
                squish.waitForObject(PublicLinkDialog.EXPIRATION_DATE_FIELD)
            )
            # Move the cursor to year (last) field and enter the year
            squish.nativeType("<Ctrl+Right>")
            squish.nativeType("<Ctrl+Right>")
            squish.nativeType(exp_year)

            # Move the cursor to day (middle) field and enter the day
            squish.nativeType("<Ctrl+Left>")
            squish.nativeType(exp_date.day)

            # Move the cursor to month (first) field and enter the month
            # Backspace works most of the time, so we clear the data using backspace
            squish.nativeType("<Ctrl+Left>")
            squish.nativeType("<Ctrl+Left>")
            squish.nativeType("<Right>")
            squish.nativeType("<Backspace>")
            squish.nativeType("<Backspace>")
            squish.nativeType(exp_date.month)
            squish.nativeType("<Return>")

            actual_date = str(
                squish.waitForObjectExists(
                    PublicLinkDialog.EXPIRATION_DATE_FIELD
                ).displayText
            )
            expected_date = f"{exp_date.month}/{exp_date.day}/{exp_year}"
            if not actual_date == expected_date:
                test.log(
                    f"Expected date: {expected_date} is not same as that of actual date: {actual_date}"
                    + ", trying workaround"
                )
                # retry with workaround
                PublicLinkDialog.set_expiration_date_with_workaround(
                    exp_year, exp_date.month, exp_date.day
                )
                actual_date = str(
                    squish.waitForObjectExists(
                        PublicLinkDialog.EXPIRATION_DATE_FIELD
                    ).displayText
                )
                if not actual_date == expected_date:
                    test.fail(
                        f"workaround failed:\nactual date: {actual_date}\nexpected date: {expected_date}"
                    )
        else:
            default_date = str(
                squish.waitForObjectExists(
                    PublicLinkDialog.EXPIRATION_DATE_FIELD
                ).displayText
            )
            PublicLinkDialog.set_default_expiry_date(default_date)

    # This workaround is needed because the above function 'set_expiration_date' can not set the month field sometime.
    # See for more details: https://github.com/owncloud/client/issues/9218
    @staticmethod
    def set_expiration_date_with_workaround(year, month, day):
        squish.mouseClick(squish.waitForObject(PublicLinkDialog.EXPIRATION_DATE_FIELD))

        # date can only be set to future date. But sometimes it can not modify the month field in first attempt.
        # date format is 'm/d/yy', so we have to edit 'month' first
        # then 'day' and then 'year'.
        # Delete data from month field
        squish.nativeType("<Delete>")
        squish.nativeType("<Delete>")
        squish.nativeType(month)
        squish.nativeType(day)
        squish.nativeType(year)
        squish.nativeType("<Return>")

    @staticmethod
    def get_radio_object_for_permssion(permissions):
        permission_locator = None
        if permissions in ("Download / View", "Viewer"):
            permission_locator = PublicLinkDialog.READ_ONLY_RADIO_BUTTON
        elif permissions in ("Download / View / Edit", "Editor"):
            permission_locator = PublicLinkDialog.READ_WRITE_RADIO_BUTTON
        elif permissions in ("Upload only (File Drop)", "Contributor"):
            permission_locator = PublicLinkDialog.UPLOAD_ONLY_RADIO_BUTTON
        else:
            raise LookupError("No such radio object found for given permission")

        return permission_locator

    @staticmethod
    def create_public_link_with_role(role):
        permission_locator = PublicLinkDialog.get_radio_object_for_permssion(role)

        squish.clickButton(squish.waitForObject(permission_locator))
        squish.clickButton(squish.waitForObject(PublicLinkDialog.CREATE_SHARE_BUTTON))
        squish.waitFor(
            lambda: (
                squish.findObject(PublicLinkDialog.PUBLIC_LINK_NAME).displayText
                == "Public link"
            )
        )

    @staticmethod
    def change_password(password):
        PublicLinkDialog.set_password(password)
        squish.clickButton(
            squish.waitForObject(PublicLinkDialog.UPDATE_PASSWORD_BUTTON)
        )

    @staticmethod
    def delete_public_link():
        squish.clickButton(squish.waitForObject(PublicLinkDialog.DELETE_LINK_BUTTON))
        squish.clickButton(
            squish.waitForObject(PublicLinkDialog.CONFIRM_LINK_DELETE_BUTTON)
        )
        squish.waitFor(
            lambda: (not object.exists(PublicLinkDialog.DELETE_LINK_BUTTON)),
        )

    @staticmethod
    def get_expiration_date():
        default_date = str(
            squish.waitForObjectExists(
                PublicLinkDialog.EXPIRATION_DATE_FIELD
            ).displayText
        )
        return PublicLinkDialog.parse_date(default_date)



import json
from types import MappingProxyType

from helpers.ConfigHelper import get_config
from helpers.api.utils import url_join
import helpers.api.http_helper as request
import helpers.api.oc10 as oc
from helpers.api.oc10 import check_success_ocs_status


share_types = MappingProxyType(
    {'user': 0, 'group': 1, 'public_link': 3, 'federated_cloud_share': 6}
)

PERMISSIONS = MappingProxyType(
    {'read': 1, 'update': 2, 'create': 4, 'delete': 8, 'share': 16}
)


def get_permission_value(permissions):
    permission_list = [perm.strip() for perm in permissions.split(',')]
    combined_permission = 0
    if 'all' in permission_list:
        return sum(PERMISSIONS.values())

    for permission in permission_list:
        if (permission_value := PERMISSIONS.get(permission)) is None:
            raise ValueError(f"Permission '{permission}' is not valid.")
        combined_permission += permission_value

    return combined_permission


def get_share_url():
    return oc.format_json(
        url_join(
            get_config('localBackendUrl'),
            'ocs',
            'v2.php',
            'apps',
            'files_sharing',
            'api',
            'v1',
            'shares',
        )
    )


def get_public_endpoint():
    return url_join(get_config('localBackendUrl'), 'remote.php', 'dav', 'public-files')


def get_public_link_shares(user):
    public_shares_list = []
    response = request.get(get_share_url(), user=user)
    check_success_ocs_status(response)
    shares = json.loads(response.text)['ocs']['data']

    for share in shares:
        if share['share_type'] == share_types['public_link']:
            public_shares_list.append(share)
    return public_shares_list


def has_public_link_share(user, resource_name):
    if public_shares := get_public_link_shares(user):
        for share in public_shares:
            if share['file_target'].strip('\\/') == resource_name:
                return True
    return False


def download_last_public_link_resource(user, resource, public_link_password=None):
    last_stime = 0
    share = False

    public_link_shares = get_public_link_shares(user)

    for share_link in public_link_shares:
        if last_stime < share_link['stime']:
            share = share_link
            last_stime = share_link['stime']
    if not share:
        raise LookupError(
            f'Expected public link for {resource} shared by {user} could not be found'
        )

    api_url = oc.format_json(url_join(get_public_endpoint(), share['token'], resource))

    response = request.get(api_url, user='public', password=public_link_password)

    if response.status_code == 200:
        return True
    if response.status_code == 404:
        return False
    raise AssertionError(f'Server returned status code: {response.status_code}')


def share_resource(user, resource, receiver, permissions, receiver_type):
    permissions = get_permission_value(permissions)
    body = {
        'path': resource,
        'shareType': share_types[receiver_type],
        'shareWith': receiver,
        'permissions': permissions,
    }
    response = request.post(get_share_url(), body, user=user)
    request.assert_http_status(
        response,
        200,
        f'Failed to share resource "{resource}" to {receiver_type} "{receiver}"',
    )
    check_success_ocs_status(response)


def create_link_share(
    user, resource, permissions, name=None, password=None, expire_date=None
):
    permissions = get_permission_value(permissions)
    body = {
        'path': resource,
        'permissions': permissions,
        'shareType': share_types['public_link'],
    }
    if name is not None:
        body['name'] = name
    if password is not None:
        body['password'] = password
    if expire_date is not None:
        body['expireDate'] = expire_date
    response = request.post(get_share_url(), body, user=user)
    request.assert_http_status(
        response, 200, f'Failed to create public link for resource "{resource}"'
    )
    check_success_ocs_status(response)


def get_shares(user):
    response = request.get(get_share_url(), user=user)
    request.assert_http_status(response, 200, 'Failed to get public link share details')
    check_success_ocs_status(response)
    return json.loads(response.text)['ocs']['data']


def get_share(user, path, share_type, share_with=None):
    shares = get_shares(user)
    for share in shares:
        if share['path'] == path and share['share_type'] == share_types[share_type]:
            if share['share_type'] == share_types['public_link']:
                return share
            if share_with and share_with == share['share_with']:
                return share
            raise ValueError(f'No share_with found for  a {share_type} share')
    raise ValueError(f'No valid share found for resource "{path}"')



import os
import re
import ctypes
import shutil

from helpers.ConfigHelper import is_windows, get_config


def build_conflicted_regex(filename):
    if "." in filename:
        # TODO: improve this for complex filenames
        namepart = filename.split(".")[0]
        extpart = filename.split(".")[1]
        # pylint: disable=anomalous-backslash-in-string
        return "%s \(conflicted copy \d{4}-\d{2}-\d{2} \d{6}\)\.%s" % (
            namepart,
            extpart,
        )
    # pylint: disable=anomalous-backslash-in-string
    return "%s \(conflicted copy \d{4}-\d{2}-\d{2} \d{6}\)" % filename


def sanitize_path(path):
    return path.replace("//", "/")


def prefix_path_namespace(path):
    if is_windows():
        # https://learn.microsoft.com/en-us/windows/win32/fileio/naming-a-file?redirectedfrom=MSDN#win32-file-namespaces
        # disable string parsing
        #  - long path
        #  - trailing whitespaces
        return f"\\\\?\\{path}"
    return path


def can_read(resource):
    read = False
    try:
        with open(resource, encoding="utf-8") as f:
            read = True
    except:
        pass
    return read and os.access(resource, os.R_OK)


def can_write(resource):
    write = False
    try:
        with open(resource, "w", encoding="utf-8") as f:
            write = True
    except:
        pass
    return write and os.access(resource, os.W_OK)


def read_file_content(file):
    with open(file, "r", encoding="utf-8") as f:
        content = f.read()
    return content


def is_empty_sync_folder(folder):
    ignore_files = ["Desktop.ini"]
    for item in os.listdir(folder):
        # do not count the hidden files as they are ignored by the client
        if not item.startswith(".") and not item in ignore_files:
            return False
    return True


def get_size_in_bytes(size):
    match = re.match(r"(\d+)((?: )?[KkMmGgBb]{0,2})?", str(size))
    units = ["b", "kb", "mb", "gb"]
    multiplier = 1024
    if match:
        size_num = int(match.group(1))
        size_unit = match.group(2)

        if not (size_unit := size_unit.lower()):
            return size_num
        if size_unit in units:
            if size_unit == "b":
                return size_num
            if size_unit == "kb":
                return size_num * multiplier
            if size_unit == "mb":
                return size_num * (multiplier**2)
            if size_unit == "gb":
                return size_num * (multiplier**3)

    raise ValueError("Invalid size: " + size)


def get_file_size_on_disk(resource_path):
    file_size_high = ctypes.c_ulonglong(0)
    if is_windows():
        return ctypes.windll.kernel32.GetCompressedFileSizeW(
            ctypes.c_wchar_p(resource_path), ctypes.pointer(file_size_high)
        )
    raise OSError("'get_file_size_on_disk' function is only supported for Windows OS.")


def get_file_size(resource_path):
    return os.stat(resource_path).st_size


# temp paths created outside the temporary directory during the test
CREATED_PATHS = []


def remember_path(path):
    CREATED_PATHS.append(path)


def cleanup_created_paths():
    global CREATED_PATHS
    for path in CREATED_PATHS:
        if os.path.exists(path):
            if os.path.isdir(path):
                shutil.rmtree(prefix_path_namespace(path))
            else:
                os.unlink(prefix_path_namespace(path))
    CREATED_PATHS = []


def get_file_for_upload(file_name):
    base_upload_dir = get_config("files_for_upload")
    return os.path.join(base_upload_dir, file_name)


def write_file(resource, content):
    with open(prefix_path_namespace(resource), "w", encoding="utf-8") as f:
        f.write(content)



import os
import subprocess
import glob
import re
from datetime import datetime
from helpers.ConfigHelper import is_windows


def get_core_dumps():
    # TODO: find a way to use coredump in windows
    if is_windows():
        return False
    # read coredump location
    with open('/proc/sys/kernel/core_pattern', 'r', encoding='utf-8') as f:
        coredump_path = f.read().strip('\n')

    # yields something like: /tmp/core-*-*-*-*
    coredump_file_pattern = re.sub(r'%[a-zA-Z]{1}', '*', coredump_path)
    return glob.glob(coredump_file_pattern)


def generate_stacktrace(scenario_title, coredumps):
    message = ['###########################################']
    message.append(f'Scenario: {scenario_title}')

    for coredump_file in coredumps:
        message.append(parse_stacktrace(coredump_file))

    message.append('###########################################')
    message.append('')
    stacktrace = '\n'.join(message)

    stacktrace_file = os.environ.get('STACKTRACE_FILE', '../stacktrace.log')
    # save stacktrace to a file
    with open(stacktrace_file, 'a', encoding='utf-8') as f:
        f.write(stacktrace)


def parse_stacktrace(coredump_file):
    message = []
    if coredump_file:
        coredump_filename = os.path.basename(coredump_file)
        # example coredump file: core-1648445754-1001-11-!drone!src!build-GUI-tests!bin!owncloud
        patterns = coredump_filename.split('-')
        app_binary = '-'.join(patterns[4:]).replace('!', '/')

        message.append('-------------------------------------------')
        message.append(f'Executable: {app_binary}')
        message.append(f'Timestamp: {str(datetime.fromtimestamp(float(patterns[1])))}')
        message.append(f'Process ID: {patterns[2]}')
        message.append(f'Signal Number: {patterns[3]}')
        message.append('-------------------------------------------')
        message.append('<<<<< STACKTRACE START >>>>>')
        message.append(
            subprocess.run(
                [
                    'gdb',
                    app_binary,
                    coredump_file,
                    '-batch',
                    '-ex',
                    'bt full',
                ],
                stdout=subprocess.PIPE,
                check=False,
            ).stdout.decode('utf-8')
        )
        message.append('<<<<< STACKTRACE END >>>>>')

        # remove coredump file
        os.unlink(coredump_file)

    return '\n'.join(message)



import names
import squish


class Settings:
    CHECKBOX_OPTION_ITEM = {
        "container": names.stack_scrollArea_QScrollArea,
        "type": "QCheckBox",
        "visible": 1,
    }
    NETWORK_OPTION_ITEM = {
        "container": names.stack_scrollArea_QScrollArea,
        "type": "QGroupBox",
        "visible": 1,
    }
    ABOUT_BUTTON = {
        "container": names.settings_stack_QStackedWidget,
        "name": "about_pushButton",
        "type": "QPushButton",
        "visible": 1,
    }
    ABOUT_DIALOG = {
        "name": "OCC__AboutDialog",
        "type": "OCC::AboutDialog",
        "visible": 1,
    }
    ABOUT_DIALOG_OK_BUTTON = {
        "text": "OK",
        "type": "QPushButton",
        "unnamed": 1,
        "visible": 1,
        "window": ABOUT_DIALOG,
    }

    GENERAL_OPTIONS_MAP = {
        "Start on Login": "autostartCheckBox",
        "Use Monochrome Icons in the system tray": "monoIconsCheckBox",
        "Language": "languageDropdown",
        "Show desktop Notifications": "desktopNotificationsCheckBox",
    }
    ADVANCED_OPTION_MAP = {
        "Sync hidden files": "syncHiddenFilesCheckBox",
        "Show crash reporter": "",
        "Edit ignored files": "ignoredFilesButton",
        "Log settings": "logSettingsButton",
        "Ask for confirmation before synchronizing folders larger than 500 MB": "newFolderLimitCheckBox",
        "Ask for confirmation before synchronizing external storages": "newExternalStorage",
    }
    NETWORK_OPTION_MAP = {
        "Proxy Settings": "proxyGroupBox",
        "Download Bandwidth": "downloadBox",
        "Upload Bandwidth": "uploadBox",
    }

    @staticmethod
    def get_checkbox_option_selector(name):
        selector = Settings.CHECKBOX_OPTION_ITEM.copy()
        selector.update({"name": name})
        if name == "languageDropdown":
            selector.update({"type": "QComboBox"})
        elif name in ("ignoredFilesButton", "logSettingsButton"):
            selector.update({"type": "QPushButton"})
        return selector

    @staticmethod
    def get_network_option_selector(name):
        selector = Settings.NETWORK_OPTION_ITEM.copy()
        selector.update({"name": name})
        return selector

    @staticmethod
    def check_general_option(option):
        selector = Settings.GENERAL_OPTIONS_MAP[option]
        squish.waitForObjectExists(Settings.get_checkbox_option_selector(selector))

    @staticmethod
    def check_advanced_option(option):
        selector = Settings.ADVANCED_OPTION_MAP[option]
        squish.waitForObjectExists(Settings.get_checkbox_option_selector(selector))

    @staticmethod
    def check_network_option(option):
        selector = Settings.NETWORK_OPTION_MAP[option]
        squish.waitForObjectExists(Settings.get_network_option_selector(selector))

    @staticmethod
    def open_about_button():
        squish.clickButton(squish.waitForObject(Settings.ABOUT_BUTTON))

    @staticmethod
    def wait_for_about_dialog_to_be_visible():
        squish.waitForObjectExists(Settings.ABOUT_DIALOG)

    @staticmethod
    def close_about_dialog():
        squish.clickButton(squish.waitForObjectExists(Settings.ABOUT_DIALOG_OK_BUTTON))



from pageObjects.AccountConnectionWizard import AccountConnectionWizard
from pageObjects.SyncConnectionWizard import SyncConnectionWizard
from pageObjects.EnterPassword import EnterPassword
from pageObjects.Toolbar import Toolbar
from pageObjects.AccountSetting import AccountSetting

from helpers.SetupClientHelper import (
    start_client,
    substitute_inline_codes,
    get_client_details,
    get_resource_path,
)
from helpers.UserHelper import get_displayname_for_user, get_password_for_user
from helpers.SyncHelper import (
    wait_for_initial_sync_to_complete,
    listen_sync_status_for_item,
)
from helpers.ConfigHelper import get_config, is_windows, is_linux, is_owncloud_client


@When('the user adds the following user credentials:')
def step(context):
    account_details = get_client_details(context.table)
    AccountConnectionWizard.add_user_credentials(
        account_details['user'], account_details['password']
    )


@Then('the account with displayname "|any|" and host "|any|" should be displayed')
def step(context, displayname, _):
    if not is_owncloud_client():
        # get displayname from first name
        displayname = get_displayname_for_user(displayname.split()[0].strip())
    displayname = substitute_inline_codes(displayname)
    Toolbar.account_exists(displayname)


@Then('the account with displayname "|any|" and host "|any|" should not be displayed')
def step(context, displayname, host):
    displayname = substitute_inline_codes(displayname)
    host = substitute_inline_codes(host)
    account_title = displayname + '\n' + host
    timeout = get_config('lowestSyncTimeout') * 1000

    test.compare(
        False,
        Toolbar.has_item(account_title, timeout),
        f"Expected account '{displayname}' to be removed",
    )


@Given('user "|any|" has set up a client with default settings')
def step(context, username):
    password = get_password_for_user(username)
    account_details = get_client_details(
        [['server', '%local_server%'], ['user', username], ['password', password]]
    )
    start_client()
    AccountConnectionWizard.add_account(account_details)
    # wait for files to sync
    wait_for_initial_sync_to_complete(get_resource_path('/', account_details['user']))


@Given('the user has set up the following accounts with default settings:')
def step(context):
    start_client()
    for idx, row in enumerate(context.table):
        username = row[0]
        password = get_password_for_user(username)
        account_details = get_client_details(
            [['server', '%local_server%'], ['user', username], ['password', password]]
        )
        if idx > 0:
            Toolbar.open_new_account_setup()
        AccountConnectionWizard.add_account(account_details)
        # wait for files to sync
        wait_for_initial_sync_to_complete(
            get_resource_path('/', account_details['user'])
        )


@Given('the user has started the client')
def step(context):
    start_client()


@When('the user starts the client')
def step(context):
    start_client()


@When('user "|any|" starts the client')
def step(context, user):
    start_client()
    listen_sync_status_for_item(get_resource_path('/', user))


@When('the user opens the add-account dialog')
def step(context):
    Toolbar.open_new_account_setup()


@When('the user adds the following account:')
def step(context):
    account_details = get_client_details(context.table)
    AccountConnectionWizard.add_account(account_details)
    # wait for files to sync
    wait_for_initial_sync_to_complete(get_resource_path('/', account_details['user']))


@Given('the user has entered the following account information:')
def step(context):
    account_details = get_client_details(context.table)
    AccountConnectionWizard.add_account_information(account_details)


@When('the user "|any|" logs out using the client-UI')
def step(context, _):
    AccountSetting.logout()


@Then('user "|any|" should be signed out')
def step(context, username):
    test.compare(
        AccountSetting.is_user_signed_out(),
        True,
        f'User "{username}" is signed out',
    )


@Given('user "|any|" has logged out from the client-UI')
def step(context, username):
    AccountSetting.logout()
    if not AccountSetting.is_user_signed_out():
        raise LookupError(f'Failed to logout user {username}')


@When('user "|any|" logs in using the client-UI')
def step(context, username):
    AccountSetting.login()
    password = get_password_for_user(username)
    enter_password = EnterPassword()
    enter_password.relogin(username, password)

    # wait for files to sync
    wait_for_initial_sync_to_complete(get_resource_path('/', username))


@When('user "|any|" logs in using the client-UI with oauth2')
def step(context, username):
    AccountSetting.login()
    password = get_password_for_user(username)
    enter_password = EnterPassword()
    enter_password.relogin(username, password, True)

    # wait for files to sync
    wait_for_initial_sync_to_complete(get_resource_path('/', username))


@When('user "|any|" opens login dialog')
def step(context, _):
    AccountSetting.login()


@When('user "|any|" enters the password "|any|"')
def step(context, username, password):
    enter_password = EnterPassword()
    enter_password.relogin(username, password)


@Then('user "|any|" should be connected to the server')
def step(context, _):
    AccountSetting.wait_until_account_is_connected()
    AccountSetting.wait_until_sync_folder_is_configured()


@When('the user removes the connection for user "|any|" and host |any|')
def step(context, username, _):
    displayname = get_displayname_for_user(username)
    displayname = substitute_inline_codes(displayname)

    Toolbar.open_account(displayname)
    AccountSetting.remove_account_connection()


@Then('connection wizard should be visible')
def step(context):
    test.compare(
        AccountConnectionWizard.is_new_connection_window_visible(),
        True,
        'Connection window is visible',
    )


@When('the user accepts the certificate')
def step(context):
    AccountConnectionWizard.accept_certificate()


@Then('the error "|any|" should be displayed in the account connection wizard')
def step(context, error_message):
    test.verify(error_message in AccountConnectionWizard.get_error_message())


@When('the user adds the server "|any|"')
def step(context, server):
    server_url = substitute_inline_codes(server)
    AccountConnectionWizard.add_server(server_url)


@When('the user selects manual sync folder option in advanced section')
def step(context):
    AccountConnectionWizard.select_manual_sync_folder_option()
    AccountConnectionWizard.next_step()


@Then('credentials wizard should be visible')
def step(context):
    test.compare(
        AccountConnectionWizard.is_credential_window_visible(),
        True,
        'Credentials wizard is visible',
    )


@When('the user selects vfs option in advanced section')
def step(context):
    AccountConnectionWizard.select_vfs_option()
    AccountConnectionWizard.next_step()


@When('the user selects download everything option in advanced section')
def step(context):
    AccountConnectionWizard.select_download_everything_option()
    AccountConnectionWizard.next_step()


@When('the user opens the advanced configuration')
def step(context):
    AccountConnectionWizard.select_advanced_config()


@Then('the user should be able to choose the local download directory')
def step(context):
    test.compare(True, AccountConnectionWizard.can_change_local_sync_dir())


@Then('the download everything option should be selected by default for Linux')
def step(context):
    if is_linux():
        test.compare(
            True,
            AccountConnectionWizard.is_sync_everything_option_checked(),
            'Sync everything option is checked',
        )


@Then('the VFS option should be selected by default for Windows')
def step(context):
    if is_windows():
        test.compare(
            True,
            AccountConnectionWizard.is_vfs_option_checked(),
            'VFS option is checked',
        )


@When(r'^the user presses the "([^"]*)" key(?:s)?', regexp=True)
def step(context, key):
    AccountSetting.press_key(key)


@Then('the log dialog should be opened')
def step(context):
    test.compare(True, AccountSetting.is_log_dialog_visible(), 'Log dialog is opened')


@When('the user adds the following oauth2 account:')
def step(context):
    account_details = get_client_details(context.table)
    account_details.update({'oauth': True})
    AccountConnectionWizard.add_account(account_details)
    # wait for files to sync
    wait_for_initial_sync_to_complete(get_resource_path('/', account_details['user']))


@Step('the user cancels the sync connection wizard')
def step(context):
    SyncConnectionWizard.cancel_folder_sync_connection_wizard()


@When('user "|any|" logs out from the login required dialog')
def step(context, _):
    enter_password = EnterPassword()
    enter_password.logout()


@When('the user quits the client')
def step(context):
    Toolbar.quit_owncloud()


@Then('"|any|" account should be opened')
def step(context, displayname):
    displayname = substitute_inline_codes(displayname)
    if not Toolbar.account_has_focus(displayname):
        raise LookupError(f"Account '{displayname}' should be opened, but it is not")


@Then(
    r'the default local sync path should contain "([^"]*)" in the (configuration|sync connection) wizard',
    regexp=True,
)
def step(context, sync_path, wizard):
    sync_path = substitute_inline_codes(sync_path)

    actual_sync_path = ''
    if wizard == 'configuration':
        actual_sync_path = AccountConnectionWizard.get_local_sync_path()
    else:
        actual_sync_path = SyncConnectionWizard.get_local_sync_path()

    test.compare(
        actual_sync_path,
        sync_path,
        'Compare sync path contains the expected path',
    )


@Then('the warning "|any|" should appear in the sync connection wizard')
def step(context, warn_message):
    actual_message = SyncConnectionWizard.get_warn_label()
    test.compare(
        True,
        warn_message in actual_message,
        'Contains warning message',
    )



import names
import squish


class AccountSetting:
    MANAGE_ACCOUNT_BUTTON = {
        "container": names.settings_dialogStack_QStackedWidget,
        "name": "manageAccountButton",
        "type": "QToolButton",
        "visible": 1,
    }
    ACCOUNT_MENU = {
        "type": "QMenu",
        "unnamed": 1,
        "visible": 1,
        "window": names.settings_OCC_SettingsDialog,
    }
    CONFIRM_REMOVE_CONNECTION_BUTTON = {
        "container": names.settings_dialogStack_QStackedWidget,
        "text": "Remove connection",
        "type": "QPushButton",
        "unnamed": 1,
        "visible": 1,
    }
    ACCOUNT_CONNECTION_LABEL = {
        "container": names.settings_dialogStack_QStackedWidget,
        "name": "connectionStatusLabel",
        "type": "QLabel",
        "visible": 1,
    }
    LOG_BROWSER_WINDOW = {
        "name": "OCC__LogBrowser",
        "type": "OCC::LogBrowser",
        "visible": 1,
    }
    ACCOUNT_LOADING = {
        "window": names.settings_OCC_SettingsDialog,
        "name": "loadingPage",
        "type": "QWidget",
        "visible": 0,
    }
    DIALOG_STACK = {
        "name": "dialogStack",
        "type": "QStackedWidget",
        "visible": 1,
        "window": names.settings_OCC_SettingsDialog,
    }
    CONFIRMATION_YES_BUTTON = {"text": "Yes", "type": "QPushButton", "visible": 1}

    @staticmethod
    def account_action(action):
        squish.clickButton(squish.waitForObject(AccountSetting.MANAGE_ACCOUNT_BUTTON))
        # wait 500ms for action menu to load
        squish.snooze(1 / 2)
        squish.activateItem(
            squish.waitForObjectItem(AccountSetting.ACCOUNT_MENU, action)
        )

    @staticmethod
    def remove_account_connection():
        AccountSetting.account_action("Remove")
        squish.clickButton(
            squish.waitForObject(AccountSetting.CONFIRM_REMOVE_CONNECTION_BUTTON)
        )

    @staticmethod
    def logout():
        AccountSetting.account_action("Log out")

    @staticmethod
    def login():
        AccountSetting.account_action("Log in")

    @staticmethod
    def get_account_connection_label():
        return str(
            squish.waitForObjectExists(AccountSetting.ACCOUNT_CONNECTION_LABEL).text
        )

    @staticmethod
    def is_connecting():
        return "Connecting to" in AccountSetting.get_account_connection_label()

    @staticmethod
    def is_user_signed_out():
        return "Signed out" in AccountSetting.get_account_connection_label()

    @staticmethod
    def is_user_signed_in():
        return "Connected" in AccountSetting.get_account_connection_label()

    @staticmethod
    def wait_until_connection_is_configured(timeout=5000):
        result = squish.waitFor(
            AccountSetting.is_connecting,
            timeout,
        )

        if not result:
            raise TimeoutError(
                "Timeout waiting for connection to be configured for "
                + str(timeout)
                + " milliseconds"
            )

    @staticmethod
    def wait_until_account_is_connected(timeout=5000):
        result = squish.waitFor(
            AccountSetting.is_user_signed_in,
            timeout,
        )

        if not result:
            raise TimeoutError(
                "Timeout waiting for the account to be connected for "
                + str(timeout)
                + " milliseconds"
            )
        return result

    @staticmethod
    def wait_until_sync_folder_is_configured(timeout=5000):
        result = squish.waitFor(
            lambda: not squish.waitForObjectExists(
                AccountSetting.ACCOUNT_LOADING
            ).visible,
            timeout,
        )

        if not result:
            raise TimeoutError(
                "Timeout waiting for sync folder to be connected for "
                + str(timeout)
                + " milliseconds"
            )
        return result

    @staticmethod
    def press_key(key):
        key = key.replace('"', "")
        key = f"<{key}>"
        squish.nativeType(key)

    @staticmethod
    def is_log_dialog_visible():
        visible = False
        try:
            visible = squish.waitForObjectExists(
                AccountSetting.LOG_BROWSER_WINDOW
            ).visible
        except:
            pass
        return visible
